/*
  container_numeric.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _CONTAINER_NUMERIC_HPP_
#define _CONTAINER_NUMERIC_HPP_
#include <boost/type_traits.hpp>
#include <boost/utility/enable_if.hpp>

#if 0
template <typename C, typename D>
C container_cast(const D &a)
{
  C b(a.size());
  copy(a.begin(), a.end(), b.begin());
  return b;
}
#else
template <typename C, typename D>
typename boost::enable_if<typename boost::is_same<C, D>, C>::type container_cast(const D &a)
{
  return a;
}
template <typename C, typename D>
typename boost::disable_if<typename boost::is_same<C, D>, C>::type container_cast(const D &a)
{
  C b(a.size());
  copy(a.begin(), a.end(), b.begin());
  return b;
}
#endif


template <typename C, typename D>
inline
C min(const C &a, const D &b)
{
  C c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = std::min(a[i], b[i]);
  return c;
}
template <typename C, typename D>
inline
C max(const C &a, const D &b)
{
  C c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = std::max(a[i], b[i]);
  return c;
}

template <typename C, typename D>
inline
C positive_min(const C &a, const D &b)
{
  C c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = (0 < a[i] && 0 < b[i]) ? std::min(a[i], b[i]) : (0 < a[i]) ? a[i] : (0 < b[i]) ? b[i] : 0;
  return c;
}
template <typename C, typename D>
inline
C negative_max(const C &a, const D &b)
{
  C c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = (a[i] < 0 && b[i] < 0) ? std::max(a[i], b[i]) : (a[i] < 0) ? a[i] : (b[i] < 0) ? b[i] : 0;
  return c;
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type min_values(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type OutputContainer;
  OutputContainer r(*first);
  InputIterator it(first);
  for (; it!=last; ++it)
    r = ::min(r, *it);
  return r;
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type max_values(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type OutputContainer;
  OutputContainer r(*first);
  InputIterator it(first);
  for (; it!=last; ++it)
    r = ::max(r, *it);
  return r;
}

template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type positive_min_values(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type OutputContainer;
  OutputContainer r(first->size());
  InputIterator it(first);
  for (; it!=last; ++it)
    r = positive_min(r, *it);
  return r;
}
template <typename InputIterator>
inline
typename std::iterator_traits<InputIterator>::value_type negative_max_values(const InputIterator first, const InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type OutputContainer;
  OutputContainer r(first->size());
  InputIterator it(first);
  for (; it!=last; ++it)
    r = negative_max(r, *it);
  return r;
}

#endif
