/*
  valarray_numeric.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
 */
#ifndef _VALARRAY_NUMERIC_HPP_
#define _VALARRAY_NUMERIC_HPP_
#include <valarray>
#include <algorithm>
#include "container_numeric.hpp"

template <typename T, typename U>
inline
std::valarray<bool> operator_less(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] < b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_less(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] < b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_less(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a < b[i];
  return c;
}

template <typename T, typename U>
inline
std::valarray<bool> operator_less_equal(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] <= b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_less_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] <= b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_less_equal(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a <= b[i];
  return c;
}

template <typename T, typename U>
inline
std::valarray<bool> operator_greater(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] > b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_greater(const std::valarray<T> &a, const std::valarray<U> &b)
{
  return operator_less(b, a);
}
template <typename T, typename U>
inline
std::valarray<bool> operator_greater(const U a, const std::valarray<T> &b)
{
  return operator_less(b, a);
}
template <typename T, typename U>
inline
std::valarray<bool> operator_greater_equal(const std::valarray<T> &a, const U b)
{
  return operator_less_equal(b, a);
}
template <typename T, typename U>
inline
std::valarray<bool> operator_greater_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
  return operator_less_equal(b, a);
}
template <typename T, typename U>
inline
std::valarray<bool> operator_greater_equal(const U a, const std::valarray<T> &b)
{
  return operator_less_equal(b, a);
}

template <typename T, typename U>
inline
std::valarray<bool> operator_equal_to(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] == b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] == b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_equal_to(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a == b[i];
  return c;
}

template <typename T, typename U>
inline
std::valarray<bool> operator_not_equal_to(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] != b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_not_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] != b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_not_equal_to(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a != b[i];
  return c;
}

#if !0 /* not unnecessary */
template <typename T, typename U>
inline
std::valarray<bool> operator_logical_and(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] && b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_logical_and(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] && b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_logical_and(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a && b[i];
  return c;
}

template <typename T, typename U>
inline
std::valarray<bool> operator_logical_or(const std::valarray<T> &a, const U b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] || b;
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_logical_or(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<bool> c(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    c[i] = a[i] || b[i];
  return c;
}
template <typename T, typename U>
inline
std::valarray<bool> operator_logical_or(const U a, const std::valarray<T> &b)
{
  std::valarray<bool> c(b.size());
  for (std::size_t i=0; i<b.size(); ++i)
    c[i] = a || b[i];
  return c;
}
#endif

template <typename T, typename U>
inline
const std::valarray<T> &operator_conditional(const U a, const std::valarray<T> &b, const std::valarray<T> &c)
{
  return a ? b : c;
}
template <typename T, typename U>
inline
std::valarray<T> operator_conditional(const std::valarray<U> &a, const std::valarray<T> &b, const std::valarray<T> &c)
{
  std::valarray<T> d(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    d[i] = a[i] ? b[i] : c[i];
  return d;
}
template <typename T, typename U, typename V>
inline
std::valarray<T> operator_conditional(const std::valarray<U> &a, const std::valarray<T> &b, const V c)
{
  std::valarray<T> d(a.size());
  for (std::size_t i=0; i<a.size(); ++i)
    d[i] = a[i] ? b[i] : c;
  return d;
}

template <typename T>
inline
T norm2(const std::valarray<T> &a)
{
  T c(0);
  for (std::size_t i=0; i<a.size(); ++i)
    c += a[i] * a[i];
  return sqrt(c);
}

template <typename T, typename U>
inline
T norm(const std::valarray<T> &a, const U b = U(2))
{
  T c(0);
  if (b != U(0)) {
    for (std::size_t i=0; i<a.size(); ++i)
      c += pow(std::abs(a[i]), b);
    c = pow(c, U(1)/b);
  }
  else { // isinf(b)
    c = std::abs(a[0]);
    for (std::size_t i=1; i<a.size(); ++i)
      c = std::max(std::abs(a[i]), c);
  }
  return c;
}

template <typename T>
inline
T prod(const std::valarray<T> &a)
{
  T c(1);
  for (std::size_t i=0; i<a.size(); ++i)
    c *= a[i];
  return c;
}

#if 0
template <typename T>
inline
std::valarray<T> abs(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = abs(c[i]);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> pow(const std::valarray<T> &a, const U b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = pow(c[i], b);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> pow(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = pow(c[i], b[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> sqrt(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = sqrt(c[i]);
  return c;
}
#endif
template <typename T>
inline
std::valarray<T> cbrt(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = cbrt(c[i]);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> hypot(const std::valarray<T> &a, const U b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = hypot(c[i], b);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> hypot(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = hypot(c[i], b[i]);
  return c;
}
#if 0
template <typename T>
inline
std::valarray<T> exp(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = exp(c[i]);
  return c;
}
#endif
template <typename T>
inline
std::valarray<T> exp2(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = exp2(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> expm1(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = expm1(c[i]);
  return c;
}
#if 0
template <typename T>
inline
std::valarray<T> log(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = log(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> log10(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = log10(c[i]);
  return c;
}
#endif
template <typename T>
inline
std::valarray<T> log2(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = log2(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> log1p(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = log1p(c[i]);
  return c;
}
#if 0
template <typename T>
inline
std::valarray<T> sin(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = sin(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> cos(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = cos(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> tan(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = tan(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> sinh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = sinh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> cosh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = cosh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> tanh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = tanh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> asin(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = asin(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> acos(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = acos(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> atan(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = atan(c[i]);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> atan2(const std::valarray<T> &a, const U b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = atan2(c[i], b);
  return c;
}
template <typename T, typename U>
inline
std::valarray<T> atan2(const std::valarray<T> &a, const std::valarray<U> &b)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = atan2(c[i], b[i]);
  return c;
}
#endif
template <typename T>
inline
std::valarray<T> asinh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = asinh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> acosh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = acosh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> atanh(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = atanh(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> erf(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = erf(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> erfc(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = erfc(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> tgamma(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = tgamma(c[i]);
  return c;
}
template <typename T>
inline
std::valarray<T> lgamma(const std::valarray<T> &a)
{
  std::valarray<T> c(a);
  for (std::size_t i=0; i<c.size(); ++i)
    c[i] = lgamma(c[i]);
  return c;
}

#endif
