/*
  valarray_numeric_quantified_operators.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _VALARRAY_NUMERIC_QUANTIFIED_OPERATORS_HPP_
#define _VALARRAY_NUMERIC_QUANTIFIED_OPERATORS_HPP_

template <typename T, typename U>
inline
bool forall_operator_less(const std::valarray<T> &a, const U b)
{
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] < b))
      return false;
  return true;
}
template <typename T, typename U>
inline
bool exists_operator_less(const std::valarray<T> &a, const U b)
{
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] < b)
      return true;
  return false;
}
template <typename T, typename U>
inline
bool forall_operator_less(const std::valarray<T> &a, const std::valarray<U> &b)
{
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] < b[i]))
      return false;
  return true;
}
template <typename T, typename U>
inline
bool exists_operator_less(const std::valarray<T> &a, const std::valarray<U> &b)
{
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] < b[i])
      return true;
  return false;
}
template <typename T, typename U>
inline
bool forall_operator_less(const U a, const std::valarray<T> &b)
{
  for (size_t i=0; i<b.size(); ++i)
    if (!(a < b[i]))
      return false;
  return true;
}
template <typename T, typename U>
inline
bool exists_operator_less(const U a, const std::valarray<T> &b)
{
  for (size_t i=0; i<b.size(); ++i)
    if (a < b[i])
      return true;
  return false;
}

template <typename T, typename U>
inline
bool forall_operator_less_equal(const std::valarray<T> &a, const U b)
{
#if 0
  return !exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] <= b))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_less_equal(const std::valarray<T> &a, const U b)
{
#if 0
  return !forall_operator_less(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] <= b)
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_less_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return !exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] <= b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_less_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return !forall_operator_less(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] <= b[i])
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_less_equal(const U a, const std::valarray<T> &b)
{
#if 0
  return !exists_operator_less(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (!(a <= b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_less_equal(const U a, const std::valarray<T> &b)
{
#if 0
  return !forall_operator_less(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (a <= b[i])
      return true;
  return false;
#endif
}

template <typename T, typename U>
inline
bool forall_operator_greater(const std::valarray<T> &a, const U b)
{
#if 0
  return forall_operator_less(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] > b))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater(const std::valarray<T> &a, const U b)
{
#if 0
  return exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] > b)
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_greater(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return forall_operator_less(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] > b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] > b[i])
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_greater(const U a, const std::valarray<T> &b)
{
#if 0
  return forall_operator_less(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (!(a > b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater(const U a, const std::valarray<T> &b)
{
#if 0
  return exists_operator_less(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (a > b[i])
      return true;
  return false;
#endif
}

template <typename T, typename U>
inline
bool forall_operator_greater_equal(const std::valarray<T> &a, const U b)
{
#if 0
  return forall_operator_less_equal(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] >= b))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater_equal(const std::valarray<T> &a, const U b)
{
#if 0
  return exists_operator_less_equal(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] >= b)
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_greater_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return forall_operator_less_equal(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] >= b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater_equal(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return exists_operator_less_equal(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] >= b[i])
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_greater_equal(const U a, const std::valarray<T> &b)
{
#if 0
  return forall_operator_less_equal(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (!(a >= b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_greater_equal(const U a, const std::valarray<T> &b)
{
#if 0
  return exists_operator_less_equal(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (a >= b[i])
      return true;
  return false;
#endif
}

template <typename T, typename U>
inline
bool forall_operator_equal_to(const std::valarray<T> &a, const U b)
{
#if 0
  return !exists_operator_less(a, b) && !exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] == b))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_equal_to(const std::valarray<T> &a, const U b)
{
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] == b)
      return true;
  return false;
}
template <typename T, typename U>
inline
bool forall_operator_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return !exists_operator_less(a, b) && !exists_operator_less(b, a);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] == b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] == b[i])
      return true;
  return false;
}
template <typename T, typename U>
inline
bool forall_operator_equal_to(const U a, const std::valarray<T> &b)
{
#if 0
  return !exists_operator_less(a, b) && !exists_operator_less(b, a);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (!(a == b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_equal_to(const U a, const std::valarray<T> &b)
{
  for (size_t i=0; i<b.size(); ++i)
    if (a == b[i])
      return true;
  return false;
}

template <typename T, typename U>
inline
bool forall_operator_not_equal_to(const std::valarray<T> &a, const U b)
{
#if 0
  return !exists_operator_equal_to(a, b);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (!(a[i] != b))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_not_equal_to(const std::valarray<T> &a, const U b)
{
#if 0
  return exists_operator_less(a, b) || exists_operator_less(b, a);
#elif 0
  return !forall_operator_equal_to(a, b);
#else
  for (size_t i=0; i<a.size(); ++i)
    if (a[i] != b)
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_not_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return !exists_operator_equal_to(a, b);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (!(a[i] != b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_not_equal_to(const std::valarray<T> &a, const std::valarray<U> &b)
{
#if 0
  return exists_operator_less(a, b) || exists_operator_less(b, a);
#elif 0
  return !forall_operator_equal_to(a, b);
#else
  for (size_t i=0; i<a.size() && i<b.size(); ++i)
    if (a[i] != b[i])
      return true;
  return false;
#endif
}
template <typename T, typename U>
inline
bool forall_operator_not_equal_to(const U a, const std::valarray<T> &b)
{
#if 0
  return !exists_operator_equal_to(a, b);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (!(a != b[i]))
      return false;
  return true;
#endif
}
template <typename T, typename U>
inline
bool exists_operator_not_equal_to(const U a, const std::valarray<T> &b)
{
#if 0
  return exists_operator_less(a, b) || exists_operator_less(b, a);
#elif 0
  return !forall_operator_equal_to(a, b);
#else
  for (size_t i=0; i<b.size(); ++i)
    if (a != b[i])
      return true;
  return false;
#endif
}

#endif
